// Project Sword & Gun Copyright © 2021 Kevin Poretti

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/CharacterMovementComponent.h"
#include "SNGCharacterMovementComponent.generated.h"

/**
 *	Class that extends UE CharacterMovementComponent to add SNG specific functionality
 *
 *	Allows for different gravity scales, friction, and walk speeds when performing certain actions. This class will also
 *	lerp and changed values 
 */
UCLASS()
class SWORDNGUN_API USNGCharacterMovementComponent : public UCharacterMovementComponent
{
	GENERATED_BODY()

public:
	USNGCharacterMovementComponent();

	virtual void PostLoad() override;
	
	/**
	 *	Instantly sets all movement parameters to their defaults values (i.e. ground friction to DefaultGroundFriction,
	 *	MaxWalkSpeed to DefaultMaxWalkSpeed, etc.)
	 */
	UFUNCTION(BlueprintCallable)
	void RestoreMovementDefaults();

	UFUNCTION(BlueprintCallable)
	void Dash();
	
	/**
	 *	Gets bool whether the character is currently in the sprinting state
	 */
	UFUNCTION(BlueprintCallable)
	FORCEINLINE bool GetIsSprinting() { return bIsSprinting; }
	
#if WITH_EDITOR
	virtual void PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent) override;
#endif // WITH_EDITOR

protected:
	// default ground 
	/** Default friction when not performing any special actions and on the ground */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	float DefaultGroundFriction;

	/** Ground friction we want character to interpolate to. Interpolated based on GroundFrictionInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetGroundFriction;

	/** How fast current GroundFriction will be interpolated to TargetGroundFriction */
	UPROPERTY(BlueprintReadWrite)
	float GroundFrictionInterpSpeed;

	/** Timer used to determine when to start interpolating GroundFriction back to TargetGroundFriction */
	UPROPERTY(BlueprintReadWrite)
	float SlideTime;
	
	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	float DefaultMaxWalkSpeed;

	/** Current MaxWalkSpeed we want to character to walk at. Interpolated based on MaxWalkSpeedInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetMaxWalkSpeed;

	UPROPERTY(BlueprintReadWrite)
	/** How fast current MaxWalkSpeed will be interpolated to TargetMaxWalkSpeed */
	float MaxWalkSpeedInterpSpeed;
	
	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Ground", meta=(ClampMin="0", UIMin="0"))
	FRotator DefaultRotationRate;
	
	// sprinting 
	/** Ground walk speed when sprinting */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float SprintingMaxWalkSpeed;

	/** Lateral speed the player needs to maintain to stay in the sprinting state */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float SprintLateralSpeedThreshold;

	/** Default ground walk speed when not performing any special actions */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	FRotator SprintingRotationRate;
	
	/** How long after applying movement input should the character start sprinting? */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float TimeToInitiateSprint;

	/** The speed in which the max walk speed to interpolated to the max sprint speed when a sprint has been initiated */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Sprint", meta=(ClampMin="0", UIMin="0"))
	float WalkToSprintInterpSpeed;
	
	FTimerHandle TimerHandle_SprintTimer;

	UPROPERTY(BlueprintReadOnly)
	bool bIsSprinting;
	
	void OnSprintTimer();

	// firing
	/** Ground walk speed when firing a ranged weapon */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Firing", meta=(ClampMin="0", UIMin="0"))
	float FiringMaxWalkSpeed;

	// reloading
	/** Ground walk speed when reloading a ranged weapon */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Reloading", meta=(ClampMin="0", UIMin="0"))
	float ReloadingMaxWalkSpeed;

	// airborne
	/** Gravity scale when not performing any special actions */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float DefaultGravityScale;

	/** How much to dampen lateral velocities at the start of a jump */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float LateralVelocityDampening;

	/** Gravity Scale we want character to interpolate to. Interpolated based on GravityScaleInterpSpeed. */
	UPROPERTY(BlueprintReadWrite)
	float TargetGravityScale;

	/** How fast current GravityScale will be interpolated to TargetGravityScale */
	UPROPERTY(BlueprintReadWrite)
	float GravityScaleInterpSpeed;

	/** Timer used to determine when to start interpolating GravityScale back to TargetGravityScale */
	UPROPERTY(BlueprintReadWrite)
	float FloatTime;
	
	/**
	 *	How high (Z) is the character able to reach when holding the jump button for the maximum allowed time.
	 *
	 *	This is used in conjunction with DesiredJumpDuration to derive DefaultGravityScale and JumpZVelocity.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MaxJumpHeight;

	/**
	 *	How long the character should be in the air for when performing a single jump and holding the jump button to reach
	 *	the maximum jump height.
	 *
	 *	This is used in conjunction with MaxJumpHeight to derive DefaultGravityScale and JumpZVelocity.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float DesiredJumpDuration;
	
	/** Force to continuously apply to character when they are holding the jump button */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float DefaultJumpZVelocity;
	
	/** Minimum height a character must reach before jump force is no longer applied even if the jump input is not being held */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MinJumpHeight;

	/** Max time the jump button can be held to give character upward velocity */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Jump", meta=(ClampMin="0", UIMin="0"))
	float MaxJumpTime;

	void ComputeGravityScaleAndJumpZVelocity();
	
	/**
	 *	Sets up character velocities and jump parameters like the jump timer	
	 */
	UFUNCTION(BlueprintCallable)
	void StartJump();

	/**
	 *	Continuously applies an upward force to the character as the jump button is held down 
	 */
	UFUNCTION(BlueprintCallable)
	void AddJumpForce();

	/** Default friction when not performing any special actions and in the air */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float DefaultAirControl;

	/** Minimum distance from the ground a character has to be to be able to perform an air dash */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float MinAllowedDashHeight;

	/** Minimum distance from the ground a character has to be to be able to perform an air attack */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Defaults|Air", meta=(ClampMin="0", UIMin="0"))
	float MinAllowedAttackHeight;

	/** Initial lateral velocity when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashVelocity;

	/** Initial ground friction when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGroundFriction;

	/** Initial gravity scale when the character dashes */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGravityScale;

	/** How long the character should "float" before the gravity scale is interpolated back to default */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashFloatTime;

	/** How fast default gravity scale will be interpolated to after dash is done */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGravityToDefaultSpeed;
	
	/** How long the character should "slide" before the ground friction is interpolated back to default */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashSlideTime;

	/** How fast default ground friction will be interpolated to after dash is done */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="SNG Character Movement|Dash", meta=(ClampMin="0", UIMin="0"))
	float DashGroundFrictionToDefaultSpeed;
	
	/**
	 * @brief Interpolates current movement parameters (i.e. GravityScale, MaxWalkSpeed, GroundFrinction) to the values
	 * specified by the parameters corresponding target value (i.e. GravityScale will be interpolated to TargetGravityScale).
	 * The interpolation speed is determined by the parameters corresponding speed value (i.e. GravityScale is interpolated
	 * to TargetGravityScale by a speed of GravityScaleInterpSpeed).
	 * 
	 * This function is called every component tick.
	 */
	void InterpolateMovementParameters(float DeltaTime);
	
	virtual void BeginPlay() override;

	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;
};
